# Copyright (c) 2021 John Heintz, Gist Labs https://gistlabs.com
# License Apache v2 http://www.apache.org/licenses/

"""


A bit of naming: vfunc is a versionedfunction, and vfuncv is a version of a versionedfunction
"""


def versionedfunction(vfunc):
    versionInfo = VersionInfo(vfunc)

    def version(vfuncv):
        versionInfo.addVersion(vfuncv)
        return vfuncv

    def default(vfuncv):
        versionInfo.setDefault(vfuncv)
        return vfuncv

    def vfunc_wrapper(*args, **kwargs):
        v = versionContext.lookupVersion(versionInfo.name)
        vfuncv = versionInfo.lookupFunction(v)
        return vfuncv(*args, **kwargs)

    vfunc_wrapper.versionInfo = versionInfo
    vfunc_wrapper.version = version
    vfunc_wrapper.default = default
    versionContext.register(versionInfo)

    return vfunc_wrapper

class VersionInfo():
    """
    This is used for each versionedfunction and connects the initial func and each version together
    """
    def __init__(self, vfunc):
        self.vfunc = vfunc
        self.versions = {}
        self.defaultVersion = None

    @property
    def name(self):
        return functionNameFrom(self.vfunc)

    def lookupFunction(self, v:str):
        if v: # some version is specified
            if v in self.versions:
                return self.versions[v]
            else:
                raise NameError(f'Version {v} not defined')
        else:
            if self.defaultVersion:
                return self.versions[self.defaultVersion]
            else:
                return self.vfunc

    def addVersion(self, vfuncv):
        versionName = versionFrom(self.vfunc.__name__, vfuncv.__name__)
        self[versionName] = vfuncv
        vfuncv.versionInfo = self

    def setDefault(self, vfuncv):
        self.defaultVersion = self.versionName(vfuncv)

    def __setitem__(self, key, value):
        self.versions[key] = value

    def versionName(self, vfuncv):
        return versionFrom(self.vfunc.__name__, vfuncv.__name__)

class VersionContext():
    """
    Global context to hold mapping from name to function to which version to use
    """
    def __init__(self):
        self.name2version = {}
        self.name2versionInfo = {} # populated during import/decorators

    def register(self, versionInfo):
        if versionInfo.name in self.name2versionInfo:
            raise NameError(f"Already registered function {versionInfo.name} in {self.name2versionInfo[versionInfo.name]}")
        self.name2versionInfo[versionInfo.name] = versionInfo

    def __getitem__(self, name):
        return self.name2version[name]

    def lookupVersion(self, name):
        if name in self.name2version:
            return self.name2version[name]
        else:
            return None

    def __setitem__(self, name, version):
        self.name2version[name] = version

versionContext = VersionContext() # versions to use for versionedfunctions, global context

def versionFrom(vfuncName, vfuncvName):
    """
    Remove the base versionedfunction name and left strip _ characters

    :param vfuncName: A versionedfunction name (string)
    :param vfuncvName: A function that is a version of a versionedfunction (name, string again)
    :return:
    """
    assert vfuncvName.startswith(vfuncName)
    return vfuncvName[len(vfuncName):].lstrip('_')

def functionNameFrom(vfunc):
    """
    The string used to identify a versionedfunction is defined by:
    * is the last two components of vfunc.__qualname__ [via split('.')]
    * if only 1 component, the prefix by module name of defining module

    class Foo():
        @versionedfunction
        def bar(self):
            pass
    would have 'Foo.bar" as __qualname__ and be used here to identify and map to versions

    <module_foo.py>
    @versionedfunction
    def bar():
        pass
    would have 'module_foo.bar' as name used to identify and map to versions

    This is intended to be a reasonable blend between fully qualified pathnames and only function name.
    """
    components = vfunc.__qualname__.split('.')[-2:] # last two components of name

    if len(components)<2:
        module = vfunc.__module__.split('.')[-1] # last module name
        components.insert(0, module)

    return '.'.join(components)


# Copyright (c) 2021 John Heintz, Gist Labs https://gistlabs.com
# License Apache v2 http://www.apache.org/licenses/

"""

"""

from versionedfunction import versionedfunction, versionContext, versionFrom
import pytest


@versionedfunction
def fooAlgo():
    return 0

def test_decorator():
    assert fooAlgo() == 0

@fooAlgo.version
def fooAlgo_v1():
    return 1

def test_fooAlgo_version_annotation():
    return fooAlgo_v1 == 1

def test_fooAlgo_versionInfo():
    assert fooAlgo.versionInfo.name == 'versionedfunction_test.fooAlgo'

def test_fooAlgo_v1_in_versions():
    assert fooAlgo.versionInfo is not None
    assert fooAlgo.versionInfo.lookupFunction('v1') == fooAlgo_v1

@fooAlgo.version
def fooAlgo_v2():
    return 2

def test_fooAlgo_v2_in_versions():
    assert fooAlgo.versionInfo.lookupFunction('v2') == fooAlgo_v2

def test_call_v2():
    versionContext['versionedfunction_test.fooAlgo'] = 'v2'
    assert fooAlgo() == 2

@versionedfunction
def barAlgo(a, b):
    return barAlgoV1(a, b)

@barAlgo.version
def barAlgoV1(a, b):
    return a + b + 1

def test_barAlgo():
    assert barAlgoV1(1, 2) == 4
    assert barAlgo(1, 2) == 4

@barAlgo.version
def barAlgoV2(a, b):
    return a + b + 2

def test_change_version():
    versionContext['versionedfunction_test.barAlgo'] = None
    assert barAlgo(1, 1) == 3
    versionContext['versionedfunction_test.barAlgo'] = 'V1'
    assert barAlgo(1, 1) == 3 # same as before
    versionContext['versionedfunction_test.barAlgo'] = 'V2'
    assert barAlgo(1, 1) == 4 # now use other version
    versionContext['versionedfunction_test.barAlgo'] = None
    assert barAlgo(1, 1) == 3 # back to default

def test_bad_version_fails():
    versionContext['versionedfunction_test.barAlgo'] = None # None is ok
    assert barAlgo(1, 1) == 3
    versionContext['versionedfunction_test.barAlgo'] = '' # Empty string is ok
    assert barAlgo(1, 1) == 3
    with pytest.raises(NameError, match="Version xyz not defined"):
        versionContext['versionedfunction_test.barAlgo'] = 'xyz'  # NOT ok
        assert barAlgo(1, 1) == 3

def test_registered_names():
    assert versionContext.name2versionInfo['versionedfunction_test.barAlgo'] == barAlgo.versionInfo
    assert versionContext.name2versionInfo['versionedfunction_test.fooAlgo'] == fooAlgo.versionInfo

def test_not_double_register():
    with pytest.raises(NameError, match="Already registered function versionedfunction_test.barAlgo"):
        class versionedfunction_test():
            @versionedfunction
            def barAlgo(self):
                pass
        assert versionedfunction_test.barAlgo.versionInfo.name == "versionedfunction_test.barAlgo"

def test_func2name():
    assert versionedfunction.__module__ == "versionedfunction.versionedfunction"
    assert versionedfunction.__qualname__ == 'versionedfunction'
    assert versionFrom.__qualname__ == 'versionFrom'

def test_method2name():
    assert versionContext.register.__qualname__ == 'VersionContext.register'
    assert versionContext.register.__module__ == "versionedfunction.versionedfunction"

def test_nestedModuleName():
    assert 'email.mime.text'.split('.')[-1] == 'text'
    assert 'text'.split('.')[-1] == 'text'

def test_listSlice():
    x = (1, 2, 3)
    assert x[-2:] == (2, 3)

    y = (3,)
    assert y[-2:] == (3,)

class Foo():
    @versionedfunction
    def algo(self):
        return 0

    @algo.version
    def algo1(self):
        return 1

    @algo.default
    @algo.version
    def algo2(self):
        return 2
foo = Foo()

def test_classmethod_versioned():
    assert foo.algo.versionInfo.name == "Foo.algo"

    assert foo.algo() == 2

    versionContext['Foo.algo'] = "1"
    assert foo.algo() == 1

def test_each_versionInfo():
    assert foo.algo1.versionInfo.name == 'Foo.algo'
    assert foo.algo.versionInfo == foo.algo1.versionInfo
    assert foo.algo1.versionInfo == foo.algo2.versionInfo

def test_default():
    assert foo.algo.versionInfo.defaultVersion == '2'
    versionContext[foo.algo.versionInfo.name] = None # remove any set values
    assert foo.algo() == 2 # must use default


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['versionedfunction']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'versionedfunction',
    'version': '0.7.26',
    'description': 'Sometimes you want to be able to dynamically call different versions of a function',
    'long_description': 'None',
    'author': 'John Heintz',
    'author_email': 'john@gistlabs.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
